@@ -1,28 +1,31 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import hashlib
+from collections import defaultdict, namedtuple
 from decimal import Decimal
-from collections import namedtuple, defaultdict
 from itertools import groupby
 
 from sql import Null
-from sql.conditionals import Coalesce
 from sql.aggregate import Max, Sum
+from sql.conditionals import Coalesce
+from sql.functions import CharLength
+from sql.operators import Concat
 
 from trytond.config import config
 from trytond.i18n import gettext
-from trytond.model import Workflow, ModelView, ModelSQL, fields, Check, \
-    sequence_ordered, DictSchemaMixin
+from trytond.model import (
+    DictSchemaMixin, Index, ModelSQL, ModelView, Workflow, fields,
+    sequence_ordered)
 from trytond.model.exceptions import AccessError
-from trytond.pyson import Eval, If, Bool
-from trytond.transaction import Transaction
-from trytond.pool import Pool
-from trytond.rpc import RPC
-from trytond.wizard import Wizard, StateView, StateAction, Button
 from trytond.modules.company import CompanyReport
+from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool
+from trytond.pyson import Bool, Eval, If
+from trytond.rpc import RPC
+from trytond.transaction import Transaction
+from trytond.wizard import Button, StateAction, StateView, Wizard
 
 from .exceptions import (
-    StatementValidateError, StatementValidateWarning, StatementPostError)
+    StatementPostError, StatementValidateError, StatementValidateWarning)
 
 if config.getboolean('account_statement', 'filestore', default=False):
     file_id = 'origin_file_id'
@@ -51,71 +54,59 @@ class Statement(Workflow, ModelSQL, ModelView):
     __name__ = 'account.statement'
 
     _states = {'readonly': Eval('state') != 'draft'}
-    _depends = ['state']
     _balance_states = _states.copy()
     _balance_states.update({
             'invisible': ~Eval('validation', '').in_(['balance']),
             'required': Eval('validation', '').in_(['balance']),
             })
-    _balance_depends = _depends + ['validation']
     _amount_states = _states.copy()
     _amount_states.update({
             'invisible': ~Eval('validation', '').in_(['amount']),
             'required': Eval('validation', '').in_(['amount']),
             })
-    _amount_depends = _depends + ['validation']
     _number_states = _states.copy()
     _number_states.update({
             'invisible': ~Eval('validation', '').in_(['number_of_lines']),
             'required': Eval('validation', '').in_(['number_of_lines']),
             })
-    _number_depends = _depends + ['validation']
 
     name = fields.Char('Name', required=True)
     company = fields.Many2One(
-        'company.company', "Company", required=True, select=True,
-        states=_states, depends=_depends)
-    journal = fields.Many2One('account.statement.journal', 'Journal',
-        required=True, select=True,
+        'company.company', "Company", required=True, states=_states)
+    journal = fields.Many2One(
+        'account.statement.journal', "Journal", required=True,
         domain=[
             ('company', '=', Eval('company', -1)),
             ],
         states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
-            },
-        depends=['state', 'company'])
+            })
     currency = fields.Function(fields.Many2One(
             'currency.currency', "Currency"), 'on_change_with_currency')
-    currency_digits = fields.Function(fields.Integer('Currency Digits'),
-        'on_change_with_currency_digits')
-    date = fields.Date('Date', required=True, select=True)
-    start_balance = fields.Numeric('Start Balance',
-        digits=(16, Eval('currency_digits', 2)),
-        states=_balance_states, depends=_balance_depends + ['currency_digits'])
-    end_balance = fields.Numeric('End Balance',
-        digits=(16, Eval('currency_digits', 2)),
-        states=_balance_states, depends=_balance_depends + ['currency_digits'])
-    balance = fields.Function(
-        fields.Numeric('Balance',
-            digits=(16, Eval('currency_digits', 2)),
-            states=_balance_states,
-            depends=_balance_depends + ['currency_digits']),
+    date = fields.Date("Date", required=True)
+    start_balance = Monetary(
+        "Start Balance", currency='currency', digits='currency',
+        states=_balance_states)
+    end_balance = Monetary(
+        "End Balance", currency='currency', digits='currency',
+        states=_balance_states)
+    balance = fields.Function(Monetary(
+            "Balance", currency='currency', digits='currency',
+            states=_balance_states),
         'on_change_with_balance')
-    total_amount = fields.Numeric('Total Amount',
-        digits=(16, Eval('currency_digits', 2)),
-        states=_amount_states, depends=_amount_depends + ['currency_digits'])
+    total_amount = Monetary(
+        "Total Amount", currency='currency', digits='currency',
+        states=_amount_states)
     number_of_lines = fields.Integer('Number of Lines',
-        states=_number_states, depends=_number_depends)
+        states=_number_states)
     lines = fields.One2Many('account.statement.line', 'statement',
         'Lines', states={
             'readonly': (Eval('state') != 'draft') | ~Eval('journal'),
-            },
-        depends=['state', 'journal'])
+            })
     origins = fields.One2Many('account.statement.origin', 'statement',
         "Origins", states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     origin_file = fields.Binary(
         "Origin File", readonly=True,
         file_id=file_id, store_prefix=store_prefix)
@@ -125,20 +116,31 @@ class Statement(Workflow, ModelSQL, ModelView):
             ('validated', "Validated"),
             ('cancelled', "Cancelled"),
             ('posted', "Posted"),
-            ], "State", readonly=True, select=True)
+            ], "State", readonly=True, sort=False)
     validation = fields.Function(fields.Char('Validation'),
         'on_change_with_validation')
     to_reconcile = fields.Function(
         fields.Boolean("To Reconcile"), 'get_to_reconcile')
 
-    del _states, _depends
-    del _balance_states, _balance_depends
-    del _amount_states, _amount_depends
-    del _number_states, _number_depends
+    del _states
+    del _balance_states
+    del _amount_states
+    del _number_states
 
     @classmethod
     def __setup__(cls):
         super(Statement, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(
+                    t,
+                    (t.journal, Index.Equality()),
+                    (t.date, Index.Range(order='DESC'))),
+                Index(
+                    t,
+                    (t.state, Index.Equality()),
+                    where=t.state.in_(['draft', 'validated'])),
+                })
         cls._order[0] = ('id', 'DESC')
         cls._transitions |= set((
                 ('draft', 'validated'),
@@ -182,16 +184,6 @@ class Statement(Workflow, ModelSQL, ModelView):
         sql_table = cls.__table__()
 
         super(Statement, cls).__register__(module_name)
-        table = cls.__table_handler__(module_name)
-
-        # Migration from 3.2: remove required on start/end balance
-        table.not_null_action('start_balance', action='remove')
-        table.not_null_action('end_balance', action='remove')
-
-        # Migration from 3.2: add required name
-        cursor.execute(*sql_table.update([sql_table.name],
-                [sql_table.id.cast(cls.name.sql_type().base)],
-                where=sql_table.name == Null))
 
         # Migration from 5.6: rename state cancel to cancelled
         cursor.execute(*sql_table.update(
@@ -211,14 +203,6 @@ class Statement(Workflow, ModelSQL, ModelView):
         Date = Pool().get('ir.date')
         return Date.today()
 
-    @staticmethod
-    def default_currency_digits():
-        Company = Pool().get('company.company')
-        if Transaction().context.get('company'):
-            company = Company(Transaction().context['company'])
-            return company.currency.digits
-        return 2
-
     @fields.depends('journal', 'state', 'lines')
     def on_change_journal(self):
         if not self.journal:
@@ -238,20 +222,7 @@ class Statement(Workflow, ModelSQL, ModelView):
 
     @fields.depends('journal')
     def on_change_with_currency(self, name=None):
-        if self.journal:
-            return self.journal.currency.id
-
-    @fields.depends('journal')
-    def on_change_with_currency_digits(self, name=None):
-        if self.journal:
-            return self.journal.currency.digits
-        return 2
-
-    def get_end_balance(self, name):
-        end_balance = self.start_balance
-        for line in self.lines:
-            end_balance += line.amount
-        return end_balance
+        return self.journal.currency if self.journal else None
 
     @fields.depends('start_balance', 'end_balance')
     def on_change_with_balance(self, name=None):
@@ -260,21 +231,9 @@ class Statement(Workflow, ModelSQL, ModelView):
 
     @fields.depends('origins', 'lines', 'journal', 'company')
     def on_change_origins(self):
-        if not self.journal or not self.origins or not self.company:
-            return
-        if self.journal.currency != self.company.currency:
-            return
-
-        invoices = set()
-        for line in self.lines:
-            if (getattr(line, 'invoice', None)
-                    and line.invoice.currency == self.company.currency):
-                invoices.add(line.invoice)
-        for origin in self.origins:
-            for line in origin.lines:
-                if (getattr(line, 'invoice', None)
-                        and line.invoice.currency == self.company.currency):
-                    invoices.add(line.invoice)
+        invoices = {l.invoice for l in self.lines if l.invoice}
+        invoices.update(
+            l.invoice for o in self.origins for l in o.lines if l.invoice)
         invoice_id2amount_to_pay = {}
         for invoice in invoices:
             if invoice.type == 'out':
@@ -287,7 +246,8 @@ class Statement(Workflow, ModelSQL, ModelView):
         for origin in origins:
             lines = list(origin.lines)
             for line in lines:
-                if (getattr(line, 'invoice', None) and line.id
+                if (line.invoice
+                        and line.id
                         and line.invoice.id in invoice_id2amount_to_pay):
                     amount_to_pay = invoice_id2amount_to_pay[line.invoice.id]
                     if (amount_to_pay
@@ -303,20 +263,12 @@ class Statement(Workflow, ModelSQL, ModelView):
             origin.lines = lines
         self.origins = origins
 
-    @fields.depends('lines', 'journal', 'company')
+    @fields.depends('lines')
     def on_change_lines(self):
         pool = Pool()
         Line = pool.get('account.statement.line')
-        if not self.journal or not self.lines or not self.company:
-            return
-        if self.journal.currency != self.company.currency:
-            return
 
-        invoices = set()
-        for line in self.lines:
-            if (getattr(line, 'invoice', None)
-                    and line.invoice.currency == self.company.currency):
-                invoices.add(line.invoice)
+        invoices = {l.invoice for l in self.lines if l.invoice}
         invoice_id2amount_to_pay = {}
         for invoice in invoices:
             if invoice.type == 'out':
@@ -328,13 +280,14 @@ class Statement(Workflow, ModelSQL, ModelView):
         lines = list(self.lines)
         line_offset = 0
         for index, line in enumerate(self.lines or []):
-            if getattr(line, 'invoice', None) and line.id:
+            if line.invoice and line.id:
                 if line.invoice.id not in invoice_id2amount_to_pay:
                     continue
+                if getattr(line, 'amount', None) is None:
+                    continue
                 amount_to_pay = invoice_id2amount_to_pay[line.invoice.id]
-                if (amount_to_pay
-                        and getattr(line, 'amount', None)
-                        and (line.amount >= 0) == (amount_to_pay <= 0)):
+                if ((line.amount > 0) == (amount_to_pay < 0)
+                        or not amount_to_pay):
                     if abs(line.amount) > abs(amount_to_pay):
                         new_line = Line()
                         for field_name, field in Line._fields.items():
@@ -504,15 +457,15 @@ class Statement(Workflow, ModelSQL, ModelView):
                 if l.invoice and l.invoice.state in {'cancelled', 'paid'})
 
         if paid_cancelled_invoice_lines:
-            warning_key = ('%s.statement_paid_cancelled_invoice_lines' %
-                    hashlib.md5(str(paid_cancelled_invoice_lines).encode(
-                        'utf-8')).hexdigest())
+            warning_key = Warning.format(
+                'statement_paid_cancelled_invoice_lines',
+                paid_cancelled_invoice_lines)
             if Warning.check(warning_key):
                 raise StatementValidateWarning(warning_key,
                     gettext('account_statement'
                         '.msg_statement_invoice_paid_cancelled'))
             Line.write(paid_cancelled_invoice_lines, {
-                    'invoice': None,
+                    'related_to': None,
                     })
 
         cls.create_move(statements)
@@ -521,9 +474,10 @@ class Statement(Workflow, ModelSQL, ModelView):
                 'state': 'validated',
                 })
         common_lines = [l for l in Line.search([
-                ('statement.state', '=', 'draft'),
-                ('invoice.state', 'in', ['posted', 'paid']),
-                ])
+                    ('statement.state', '=', 'draft'),
+                    ('related_to.state', 'in', ['posted', 'paid'],
+                        'account.invoice'),
+                    ])
             if l.invoice.reconciled]
         if common_lines:
             warning_key = '_'.join(str(l.id) for l in common_lines)
@@ -532,7 +486,7 @@ class Statement(Workflow, ModelSQL, ModelView):
                     gettext('account_statement'
                         '.msg_statement_paid_invoice_draft'))
             Line.write(common_lines, {
-                    'invoice': None,
+                    'related_to': None,
                     })
 
     @classmethod
@@ -571,6 +525,8 @@ class Statement(Workflow, ModelSQL, ModelView):
             amount_second_currency = 0
             for line in lines:
                 move_line = line.get_move_line()
+                if not move_line:
+                    continue
                 move_line.move = move
                 amount += move_line.debit - move_line.credit
                 if move_line.amount_second_currency:
@@ -593,14 +549,13 @@ class Statement(Workflow, ModelSQL, ModelView):
         Move = pool.get('account.move')
         Period = pool.get('account.period')
 
-        period_id = Period.find(self.company.id, date=key['date'])
+        period = Period.find(self.company, date=key['date'])
         return Move(
-            period=period_id,
+            period=period,
             journal=self.journal.journal,
             date=key['date'],
             origin=self,
             company=self.company,
-            description=str(key['number']),
             )
 
     def _get_move_line(self, amount, amount_second_currency, lines):
@@ -615,19 +570,12 @@ class Statement(Workflow, ModelSQL, ModelView):
             second_currency = None
             amount_second_currency = None
 
-        descriptions = {l.description for l in lines}
-        if len(descriptions) == 1:
-            description, = descriptions
-        else:
-            description = ''
-
         return MoveLine(
             debit=abs(amount) if amount < 0 else 0,
             credit=abs(amount) if amount > 0 else 0,
             account=self.journal.account,
             second_currency=second_currency,
             amount_second_currency=amount_second_currency,
-            description=description,
             )
 
     @classmethod
@@ -683,13 +631,12 @@ class Statement(Workflow, ModelSQL, ModelView):
         return new_statements
 
 
-def origin_mixin(_states, _depends):
+def origin_mixin(_states):
     class Mixin:
         __slots__ = ()
         statement = fields.Many2One(
             'account.statement', "Statement",
-            required=True, ondelete='CASCADE', states=_states,
-            depends=_depends)
+            required=True, ondelete='CASCADE', states=_states)
         statement_state = fields.Function(
             fields.Selection('get_statement_states', "Statement State"),
             'on_change_with_statement_state')
@@ -698,21 +645,18 @@ def origin_mixin(_states, _depends):
             'on_change_with_company', searcher='search_company')
         number = fields.Char("Number")
         date = fields.Date(
-            "Date", required=True, states=_states, depends=_depends)
-        amount = fields.Numeric(
-            "Amount", required=True,
-            digits=(16, Eval('currency_digits', 2)),
-            states=_states, depends=_depends + ['currency_digits'])
+            "Date", required=True, states=_states)
+        amount = Monetary(
+            "Amount", currency='currency', digits='currency', required=True,
+            states=_states)
         currency = fields.Function(fields.Many2One(
                 'currency.currency', "Currency"), 'on_change_with_currency')
-        currency_digits = fields.Function(fields.Integer(
-                "Currency Digits"), 'on_change_with_currency_digits')
         party = fields.Many2One(
             'party.party', "Party", states=_states,
             context={
                 'company': Eval('company', -1),
                 },
-            depends=_depends + ['company'])
+            depends={'company'})
         account = fields.Many2One(
             'account.account', "Account",
             domain=[
@@ -723,15 +667,20 @@ def origin_mixin(_states, _depends):
             context={
                 'date': Eval('date'),
                 },
-            states=_states, depends=_depends + ['company', 'date'])
-        description = fields.Char(
-            "Description", states=_states, depends=_depends)
+            states=_states, depends={'date'})
+        description = fields.Char("Description", states=_states)
 
         @classmethod
         def __setup__(cls):
+            cls.number.search_unaccented = False
             super().__setup__()
             cls.__access__.add('statement')
 
+        @classmethod
+        def order_number(cls, tables):
+            table, _ = tables[None]
+            return [CharLength(table.number), table.number]
+
         @classmethod
         def get_statement_states(cls):
             pool = Pool()
@@ -745,8 +694,7 @@ def origin_mixin(_states, _depends):
 
         @fields.depends('statement', '_parent_statement.company')
         def on_change_with_company(self, name=None):
-            if self.statement and self.statement.company:
-                return self.statement.company.id
+            return self.statement.company if self.statement else None
 
         @classmethod
         def search_company(cls, name, clause):
@@ -755,12 +703,7 @@ def origin_mixin(_states, _depends):
         @fields.depends('statement', '_parent_statement.journal')
         def on_change_with_currency(self, name=None):
             if self.statement and self.statement.journal:
-                return self.statement.journal.currency.id
-
-        @fields.depends('statement', '_parent_statement.journal')
-        def on_change_with_currency_digits(self, name=None):
-            if self.statement and self.statement.journal:
-                return self.statement.journal.currency.digits
+                return self.statement.journal.currency
 
     return Mixin
 
@@ -768,12 +711,9 @@ def origin_mixin(_states, _depends):
 _states = {
     'readonly': Eval('statement_state') != 'draft',
     }
-_depends = ['statement_state']
 
 
-class Line(
-        origin_mixin(_states, _depends), sequence_ordered(),
-        ModelSQL, ModelView):
+class Line(origin_mixin(_states), sequence_ordered(), ModelSQL, ModelView):
     'Account Statement Line'
     __name__ = 'account.statement.line'
 
@@ -782,50 +722,110 @@ class Line(
             ('company', '=', Eval('company', -1)),
             ],
         depends=['company'])
-    invoice = fields.Many2One('account.invoice', 'Invoice',
-        domain=[
-            ('company', '=', Eval('company', -1)),
-            If(Bool(Eval('party')), [('party', '=', Eval('party'))], []),
-            If(Bool(Eval('account')), [('account', '=', Eval('account'))], []),
-            If(Eval('statement_state') == 'draft',
-                ('state', '=', 'posted'),
-                ('state', '!=', '')),
-            ],
+    related_to = fields.Reference(
+        "Related To", 'get_relations',
+        domain={
+            'account.invoice': [
+                ('company', '=', Eval('company', -1)),
+                ('currency', '=', Eval('currency', -1)),
+                If(Bool(Eval('party')),
+                    ['OR',
+                        ('party', '=', Eval('party', -1)),
+                        ('alternative_payees', '=', Eval('party', -1)),
+                        ],
+                    []),
+                If(Bool(Eval('account')),
+                    ('account', '=', Eval('account')),
+                    ()),
+                If(Eval('statement_state') == 'draft',
+                    ('state', '=', 'posted'),
+                    ('state', '!=', '')),
+                ],
+            },
         states=_states,
-        context={'with_payment': False},
-        depends=['company', 'party', 'account'] + _depends)
+        context={'with_payment': False})
     origin = fields.Many2One('account.statement.origin', 'Origin',
         readonly=True,
         states={
             'invisible': ~Bool(Eval('origin')),
             },
         domain=[
-            ('statement', '=', Eval('statement')),
-            ('date', '=', Eval('date')),
-            ],
-        depends=['statement', 'date'])
+            ('statement', '=', Eval('statement', -1)),
+            ('date', '=', Eval('date', None)),
+            ])
+    party_required = fields.Function(
+        fields.Boolean("Party Required"), 'on_change_with_party_required')
 
     @classmethod
     def __setup__(cls):
         super(Line, cls).__setup__()
-        if 'origin' not in cls.date.depends:
-            cls.date.states.update({
-                    'readonly': (cls.date.states['readonly']
-                        | Bool(Eval('origin', 0))),
-                    })
-            cls.date.depends.append('origin')
+        cls.date.states = {
+            'readonly': (
+                (Eval('statement_state') != 'draft')
+                | Bool(Eval('origin', 0))),
+            }
         cls.account.required = True
-        t = cls.__table__()
-        cls._sql_constraints += [
-            ('check_statement_line_amount', Check(t, t.amount != 0),
-                'account_statement.msg_line_amount_non_zero'),
-            ]
+        cls.party.states = {
+            'required': (Eval('party_required', False)
+                & (Eval('statement_state') == 'draft')),
+            }
+
+    @classmethod
+    def __register__(cls, module):
+        table = cls.__table__()
+
+        super().__register__(module)
+
+        table_h = cls.__table_handler__(module)
+        cursor = Transaction().connection.cursor()
+
+        # Migration from 6.2: replace invoice by related_to
+        if table_h.column_exist('invoice'):
+            cursor.execute(*table.update(
+                    [table.related_to],
+                    [Concat('account.invoice,', table.invoice)],
+                    where=table.invoice != Null))
+            table_h.drop_column('invoice')
+
+        # Migration from 6.6: Allow amount of zero
+        table_h.drop_constraint('check_statement_line_amount')
+
+    @property
+    @fields.depends('related_to')
+    def invoice(self):
+        pool = Pool()
+        Invoice = pool.get('account.invoice')
+        related_to = getattr(self, 'related_to', None)
+        if isinstance(related_to, Invoice) and related_to.id >= 0:
+            return related_to
+
+    @invoice.setter
+    def invoice(self, value):
+        self.related_to = value
+
+    @fields.depends('account')
+    def on_change_with_party_required(self, name=None):
+        if self.account:
+            return self.account.party_required
+        return False
 
     @staticmethod
     def default_amount():
         return Decimal(0)
 
-    @fields.depends('amount', 'party', 'invoice', 'date')
+    @classmethod
+    def _get_relations(cls):
+        "Return a list of Model names for related_to Reference"
+        return ['account.invoice']
+
+    @classmethod
+    def get_relations(cls):
+        Model = Pool().get('ir.model')
+        get_name = Model.get_name
+        models = cls._get_relations()
+        return [(None, '')] + [(m, get_name(m)) for m in models]
+
+    @fields.depends('amount', 'party', 'date', methods=['invoice'])
     def on_change_party(self):
         if self.party:
             if self.amount:
@@ -837,15 +837,17 @@ class Line(
 
         if self.invoice:
             if self.party:
-                if self.invoice.party != self.party:
+                if (self.invoice.party != self.party
+                        or self.party not in self.invoice.alternative_payees):
                     self.invoice = None
             else:
                 self.invoice = None
 
-    @fields.depends('amount', 'party', 'account', 'invoice', 'statement',
-        'date', '_parent_statement.journal')
+    @fields.depends(
+        'amount', 'party', 'account', 'date',
+        'statement', '_parent_statement.journal',
+        methods=['invoice'])
     def on_change_amount(self):
-        Currency = Pool().get('currency.currency')
         if self.party:
             with Transaction().set_context(date=self.date):
                 if self.account and self.account not in (
@@ -858,19 +860,8 @@ class Line(
                         self.account = self.party.account_receivable_used
                     else:
                         self.account = self.party.account_payable_used
-        if self.invoice:
-            if self.amount and self.statement and self.statement.journal:
-                invoice = self.invoice
-                journal = self.statement.journal
-                with Transaction().set_context(date=invoice.currency_date):
-                    amount_to_pay = Currency.compute(invoice.currency,
-                        invoice.amount_to_pay, journal.currency)
-                if abs(self.amount) > amount_to_pay:
-                    self.invoice = None
-            else:
-                self.invoice = None
 
-    @fields.depends('account', 'invoice')
+    @fields.depends('account', methods=['invoice'])
     def on_change_account(self):
         if self.invoice:
             if self.account:
@@ -879,11 +870,17 @@ class Line(
             else:
                 self.invoice = None
 
-    @fields.depends('party', 'account', 'invoice')
-    def on_change_invoice(self):
+    @fields.depends('party', 'account', methods=['invoice'])
+    def on_change_related_to(self):
         if self.invoice:
             if not self.party:
-                self.party = self.invoice.party
+                if not self.invoice.alternative_payees:
+                    self.party = self.invoice.party
+                else:
+                    try:
+                        self.party, = self.invoice.alternative_payees
+                    except ValueError:
+                        pass
             if not self.account:
                 self.account = self.invoice.account
 
@@ -909,11 +906,11 @@ class Line(
     @fields.depends('origin', '_parent_origin.company')
     def on_change_with_company(self, name=None):
         try:
-            company = super(Line, self).on_change_with_company()
+            company = super().on_change_with_company()
         except AttributeError:
             company = None
-        if self.origin and self.origin.company:
-            return self.origin.company.id
+        if self.origin and hasattr(self.origin, 'company'):
+            company = self.origin.company
         return company
 
     @fields.depends('origin', '_parent_origin.statement_state')
@@ -940,7 +937,7 @@ class Line(
         else:
             default = default.copy()
         default.setdefault('move', None)
-        default.setdefault('invoice', None)
+        default.setdefault('related_to', None)
         return super(Line, cls).copy(lines, default=default)
 
     @classmethod
@@ -964,8 +961,15 @@ class Line(
                 MoveLine.reconcile(*to_reconcile)
                 to_reconcile.clear()
 
+            if move_line.second_currency:
+                amount = -move_line.amount_second_currency
+                currency = move_line.second_currency
+            else:
+                amount = move_line.credit - move_line.debit
+                currency = line.company.currency
+
             reconcile_lines = line.invoice.get_reconcile_lines_for_amount(
-                move_line.credit - move_line.debit)
+                amount, currency, party=line.party)
 
             assert move_line.account == line.invoice.account
 
@@ -1003,6 +1007,8 @@ class Line(
         MoveLine = pool.get('account.move.line')
         Currency = Pool().get('currency.currency')
         zero = Decimal("0.0")
+        if not self.amount:
+            return
         with Transaction().set_context(date=self.date):
             amount = Currency.compute(self.statement.journal.currency,
                 self.amount, self.statement.company.currency)
@@ -1015,7 +1021,6 @@ class Line(
 
         return MoveLine(
             origin=self,
-            description=self.description,
             debit=amount < zero and -amount or zero,
             credit=amount >= zero and amount or zero,
             account=self.account,
@@ -1025,7 +1030,7 @@ class Line(
             )
 
 
-del _states, _depends
+del _states
 
 
 class LineGroup(ModelSQL, ModelView):
@@ -1037,18 +1042,16 @@ class LineGroup(ModelSQL, ModelView):
             'Journal'), 'get_journal', searcher='search_journal')
     number = fields.Char('Number')
     date = fields.Date('Date')
-    amount = fields.Numeric('Amount',
-        digits=(16, Eval('currency_digits', 2)),
-        depends=['currency_digits'])
+    amount = Monetary(
+        "Amount", currency='currency', digits='currency')
     currency = fields.Function(fields.Many2One('currency.currency',
             'Currency'), 'get_currency')
-    currency_digits = fields.Function(fields.Integer('Currency Digits'),
-        'get_currency_digits')
     party = fields.Many2One('party.party', 'Party')
     move = fields.Many2One('account.move', 'Move')
 
     @classmethod
     def __setup__(cls):
+        cls.number.search_unaccented = False
         super(LineGroup, cls).__setup__()
         cls.__access__.add('statement')
         cls._order.insert(0, ('date', 'DESC'))
@@ -1084,10 +1087,15 @@ class LineGroup(ModelSQL, ModelView):
         return move.join(line,
             condition=move.id == line.move
             ).select(*columns,
-                where=move.origin.ilike(Statement.__name__ + ',%'),
+                where=move.origin.like(Statement.__name__ + ',%'),
                 group_by=std_columns + [move.id]
                 )
 
+    @classmethod
+    def order_number(cls, tables):
+        table, _ = tables[None]
+        return [CharLength(table.number), table.number]
+
     def get_journal(self, name):
         return self.statement.journal.id
 
@@ -1098,17 +1106,13 @@ class LineGroup(ModelSQL, ModelView):
     def get_currency(self, name):
         return self.statement.journal.currency.id
 
-    def get_currency_digits(self, name):
-        return self.statement.journal.currency.digits
-
 
 _states = {
     'readonly': (Eval('statement_state') != 'draft') | Eval('lines', []),
     }
-_depends = ['statement_state']
 
 
-class Origin(origin_mixin(_states, _depends), ModelSQL, ModelView):
+class Origin(origin_mixin(_states), ModelSQL, ModelView):
     "Account Statement Origin"
     __name__ = 'account.statement.origin'
     _rec_name = 'number'
@@ -1120,16 +1124,13 @@ class Origin(origin_mixin(_states, _depends), ModelSQL, ModelView):
                 | ~Eval('statement_state').in_(['draft', 'validated'])),
             },
         domain=[
-            ('statement', '=', Eval('statement')),
-            ('date', '=', Eval('date')),
-            ],
-        depends=['statement', 'date', 'statement_id'])
+            ('statement', '=', Eval('statement', -1)),
+            ('date', '=', Eval('date', None)),
+            ])
     statement_id = fields.Function(
         fields.Integer("Statement ID"), 'on_change_with_statement_id')
-    pending_amount = fields.Function(
-        fields.Numeric("Pending Amount",
-            digits=(16, Eval('_parent_statement', {})
-                .get('currency_digits', 2))),
+    pending_amount = fields.Function(Monetary(
+            "Pending Amount", currency='currency', digits='currency'),
         'on_change_with_pending_amount', searcher='search_pending_amount')
     information = fields.Dict(
         'account.statement.origin.information', "Information", readonly=True)
@@ -1179,7 +1180,7 @@ class Origin(origin_mixin(_states, _depends), ModelSQL, ModelView):
         return super().copy(origins, default=default)
 
 
-del _states, _depends
+del _states
 
 
 class OriginInformation(DictSchemaMixin, ModelSQL, ModelView):
@@ -1192,7 +1193,8 @@ class ImportStatementStart(ModelView):
     __name__ = 'account.statement.import.start'
     company = fields.Many2One('company.company', "Company", required=True)
     file_ = fields.Binary("File", required=True)
-    file_format = fields.Selection([(None, '')], 'File Format', required=True)
+    file_format = fields.Selection(
+        [(None, '')], "File Format", required=True, translate=False)
 
     @classmethod
     def default_file_format(cls):
@@ -1220,6 +1222,7 @@ class ImportStatement(Wizard):
         for statement in statements:
             statement.origin_file = fields.Binary.cast(self.start.file_)
         Statement.save(statements)
+        self.start.file_ = None
 
         data = {'res_id': list(map(int, statements))}
         if len(statements) == 1:
